/**
 * Juliac
 * Copyright (C) 2009-2013 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Christophe Munilla
 */
package org.objectweb.fractal.juliac.osgi.revision;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.List;
import java.util.Vector;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.objectweb.fractal.juliac.osgi.concierge.Concierge;
import org.objectweb.fractal.juliac.osgi.revision.api.BundleContextClassLoader;
import org.osgi.framework.Bundle;

/***
 * BundleClassLoader implementation for Concierge Framework
 */
public class BundleContextClassLoaderImpl extends URLClassLoader implements
        BundleContextClassLoader
{
    /**
     * Logger
     */
    private static final Logger LOGGER = Logger.getLogger(
            BundleContextClassLoaderImpl.class.getCanonicalName()); 

    /**
     * The BundleContextClassLoader instance
     */
    private static BundleContextClassLoader instance;
    
    /**
     * Return the current BundleContextClassLoader instance
     * 
     * @param parent
     *          the parent ClassLoader of the BundleContextClassLoader to instantiate
     * @return  
     *          the BundleContextClassLoader instance
     */
    public static BundleContextClassLoader getBundleContextClassLoader(
            ClassLoader parent)
    {
        if(instance == null)
        {
            instance = new BundleContextClassLoaderImpl(
                    Concierge.getInstance(),parent);
        }
        return instance;
    }
    
    /**
     * Class<?> of the Concierge Framework's BundleClassLoader
     */
    private Class<?> bundleClassLoaderClass = null;   

    /**
     * Resources/Classes search methods of the Concierge 
     * Framework's BundleClassLoader
     */
    private Method findOwnResources = null; 
    private Method findDelegatedClass = null;
    private Method findClass = null;  
    
    /**
     * The associated Concierge Framework instance
     */
    private Concierge platform;
    
    /**
     * Constructor
     * 
     * @param platform
     * @param parent
     */
    protected BundleContextClassLoaderImpl(Concierge platform, 
            ClassLoader parent)
    {
        super(new URL[0],parent==null?Concierge.class.getClassLoader():parent);
        this.platform = platform;
               
        bundleClassLoaderClass = platform.getBundleClassLoader(
                getBundle(1)).getClass();
        try
        {
            findOwnResources = bundleClassLoaderClass.getDeclaredMethod("findOwnResources",
                    new Class<?>[]{ String.class , boolean.class });
            findOwnResources.setAccessible(true);
            findDelegatedClass = bundleClassLoaderClass.getDeclaredMethod("findDelegatedClass",
                    new Class<?>[]{bundleClassLoaderClass,String.class});
            findDelegatedClass.setAccessible(true);
            findClass = bundleClassLoaderClass.getDeclaredMethod("findClass",
                    new Class<?>[]{String.class});
            findClass.setAccessible(true);
            
        } catch (Throwable throwable)
        {
            if(LOGGER.isLoggable(Level.CONFIG))
            {
                LOGGER.log(Level.CONFIG,throwable.getMessage(),throwable);
            }           
        }      
    }

    /**
     * {@inheritDoc}
     * 
     * @see java.lang.ClassLoader#loadClass(java.lang.String)
     */
    public synchronized Class<?> loadClass(String className) 
        throws ClassNotFoundException
    {   
        int pos = className.lastIndexOf('.');
        String packageName =  pos > -1 ? className.substring(0, pos) : "";
        
        ClassLoader loader = platform.getBundleClassLoader(packageName);
        if(loader == null)
        {
           return getParent().loadClass(className);
        }
        Class<?> clazz = null;
        try{                
            clazz = (Class<?>) findDelegatedClass.invoke(
                    null, new Object[]{loader,className});
            
        } catch(Throwable throwable)
        {
            if(LOGGER.isLoggable(Level.CONFIG))
            {
                LOGGER.log(Level.CONFIG,throwable.getMessage(),throwable);
            }                
        }
        if(clazz == null)
        {
            clazz = getParent().loadClass(className);
        }
        return clazz;
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.objectweb.fractal.juliac.osgi.BundleContextClassLoader#
     * loadClass(java.lang.String,long)
     */
    public Class<?> loadClass(String className, long bundleId)
            throws ClassNotFoundException
    {
        Class<?> clazz = null;  
        ClassLoader classloader = platform.getBundleClassLoader(bundleId);
        try{
            
            clazz = (Class<?>) findClass.invoke(classloader,
                    new Object[]{className});
            
        } catch(Throwable throwable)
        {
            if(LOGGER.isLoggable(Level.CONFIG))
            {
                LOGGER.log(Level.CONFIG,throwable.getMessage(),throwable);
            }                
        }
        if(clazz == null)
        {   
            loadClass(className);
        }
        return clazz;
    }

    /**
     * {@inheritDoc}
     * 
     * @see java.lang.ClassLoader#getResource(java.lang.String)
     */
    public URL getResource(String resourceName)
    {
        URL resourceURL = null; 
        String formatResourceName =  resourceName.startsWith("/")?
                resourceName.substring(1):resourceName; 
        int pos = formatResourceName .lastIndexOf('/');
        String resourcePackage = pos>-1?formatResourceName.substring(0,pos).replace(
                '/', '.'):formatResourceName;
        ClassLoader loader = platform.getBundleClassLoader(resourcePackage);
        if(loader != null)
        {
            try
            {
                Vector resourcesVector  = (Vector) findOwnResources.invoke(
                        loader, new Object[]{formatResourceName, false});
                    
                if(resourcesVector != null && resourcesVector.size()>0)
                {
                    resourceURL = (URL) resourcesVector.elementAt(0);
                }
            }  catch (Throwable throwable)
            {
                if(LOGGER.isLoggable(Level.CONFIG))
                {
                    LOGGER.log(Level.CONFIG,throwable.getMessage(),throwable);
                }      
            }            
        } else
        {
            Enumeration<Bundle> bundles = platform.bundles();
            while(bundles.hasMoreElements())
            {
                try
                {
                    Vector resourcesVector  = (Vector) findOwnResources.invoke(
                        platform.getBundleClassLoader(bundles.nextElement()),
                        new Object[]{formatResourceName, false});
                    
                    if(resourcesVector != null && resourcesVector.size()>0)
                    {
                        resourceURL = (URL) resourcesVector.elementAt(0);
                        break;
                    }
                } catch (Throwable throwable)
                {
                    if(LOGGER.isLoggable(Level.CONFIG))
                    {
                        LOGGER.log(Level.CONFIG,throwable.getMessage(),throwable);
                    }      
                }
            }
        }
        return resourceURL;
    }

    /**
     * {@inheritDoc}
     * @throws IOException 
     * 
     * @see org.objectweb.fractal.juliac.osgi.BundleContextClassLoader#getResources(java.lang.String,
     *      java.util.List)
     */
    public Enumeration<URL> getResources(String resourceName) throws IOException
    {   
        List<URL> resources = new ArrayList<URL>();

        Enumeration<Bundle> bundles = platform.bundles();
        while(bundles.hasMoreElements())
        {
            try
            {
                Vector results  = (Vector) findOwnResources.invoke(
                        platform.getBundleClassLoader(bundles.nextElement()),
                        new Object[]{ resourceName,true });
                
                if(results != null)
                {                
                   resources.addAll((Vector<URL>) results);
                }
            } catch (Throwable throwable)
            {
                if(LOGGER.isLoggable(Level.CONFIG))
                {
                    LOGGER.log(Level.CONFIG,throwable.getMessage(),throwable);
                }      
            }
        }
        return Collections.enumeration(resources);
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.objectweb.fractal.juliac.osgi.revision.BundleContextClassLoader#getEntryPaths(java.lang.String,
     *      long)
     */
    public synchronized Enumeration<String> getEntryPaths(String path, long bundleId)
    {        
        final List<String> entriesName = new ArrayList<String>();   
        File bundleStore = getBundleStore(bundleId);
        String formatPath = (path.startsWith("/")?path.substring(1):path);
        
        if(bundleStore != null)
        {   
            bundleStore = new File(bundleStore,formatPath);
        
            String[] bundleStoreList = bundleStore.list();
            Arrays.sort(bundleStoreList, new Comparator<String>()
            {
                public int compare(String string1, String string2)
                {
                    return (string1.length()<string2.length()?
                            -1:(string1.length()==string2.length()?
                                    0:1));
                }        
            });
            if(bundleStoreList != null && bundleStoreList.length > 0)
            {   
                int index = 0;
                for(;index < bundleStoreList.length;index++)
                {
                    entriesName.add(formatPath + 
                            File.separatorChar +
                            bundleStoreList[index]);                    
                }
            }           
            return new Enumeration<String>()
            {
                private int position = 0;
                private final List<String> entriesList = entriesName;
                
                public boolean hasMoreElements()
                {
                    return position < entriesList.size();
                }
                public String nextElement()
                {
                    String entry = entriesList.get(position++);
                    return entry;
                }            
            };
        } else 
        {
            return Collections.enumeration(
                    Collections.<String>emptyList());
        }
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.objectweb.fractal.juliac.osgi.revision.BundleContextClassLoader#
     * getEntry(java.lang.String,long)
     */
    public URL getEntry(String path, long bundleId)
    {
        URL entryURL = null; 
        
        File bundleStore = getBundleStore(bundleId);
        if(bundleStore != null)
        {   
            File entryFile = new File(bundleStore,
                    (path.startsWith("/")?path.substring(1):path));
            
            if(entryFile.exists())
            {
                try
                {
                    entryURL = entryFile.toURI().toURL();
                    
                } catch (MalformedURLException e)
                {
                    e.printStackTrace();
                }
            }            
        }
        return entryURL;        
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.objectweb.fractal.juliac.osgi.revision.api.BundleContextClassLoader#
     * uninstallBundle(long)
     */
    public void uninstallBundle(long bundleId)
    {
        deleteAllRecursively(getBundleStore(bundleId));
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.objectweb.fractal.juliac.osgi.revision.api.BundleContextClassLoader#
     * updateBundle(org.osgi.framework.Bundle)
     */
    public void updateBundle(Bundle bundle)
    {
        uninstallBundle(bundle.getBundleId());
    }
    
    /**
     * {@inheritDoc}
     * 
     * @see org.objectweb.fractal.juliac.osgi.revision.api.BundleContextClassLoader#
     * registerBundle(org.osgi.framework.Bundle)
     */
    public void registerBundle(Bundle bundle)
    {
        //do nothing
    }
    
    /**
     * @param bundleId
     * @return
     */
    private Bundle getBundle(long bundleId)
    {
        Enumeration<Bundle> bundles = platform.bundles();
        if(bundles != null)
        {
            while(bundles.hasMoreElements())
            {
                Bundle bundle = bundles.nextElement();
                if(bundle.getBundleId() == bundleId)
                {
                    return bundle;
                }
            }
        }
        return null;
    }
    
    /**
     * @param bundleId
     * @return
     */
    private boolean isClassPathDefined(long bundleId)
    {
        Bundle bundle = getBundle(bundleId);
        if(bundle != null)
        {         
            String bundleClassPathHeader = (String) bundle.getHeaders().get(
                "Bundle-ClassPath");        
            return (bundleClassPathHeader!=null  &&
                    bundleClassPathHeader.length()>1);
        }
        return false;
    }
    
    /**
     * Retrieve or/and create the store directory where a bundle is 
     * decompressed
     * 
     * @param bundleId
     *          the identifier of the Bundle 
     * @return
     */
    private File getBundleStore(long bundleId)
    {   
        String bundleStorePath = null;
        File bundleStore = null;
        
        if(platform.decompressed() && isClassPathDefined(bundleId))
        {   
            bundleStorePath = new StringBuilder(platform.storage()).append(
                    File.separatorChar).append(bundleId).append(
                            File.separatorChar).append("content").toString();
            
            bundleStore = new File(bundleStorePath);            
        }else
        {   
            try
            {
                File tmpFile = File.createTempFile("bccl", ".tmp");
                tmpFile.delete();
                tmpFile = tmpFile.getParentFile();
                
                bundleStorePath = new StringBuilder(tmpFile.getAbsolutePath()).append(
                        File.separatorChar).append("concierge_resources").append(
                                  File.separatorChar).append(bundleId).toString();
                
                bundleStore = new File(bundleStorePath);
                if(!bundleStore.exists() && bundleStore.mkdirs())
                {
                    decompressBundle(bundleId,bundleStore);
                }
            } catch (IOException e)
            {
                e.printStackTrace();
            }
        }
        return bundleStore;
    }
    
    /**
     * @param bundle
     * @param bundleStore
     * @throws IOException 
     */
    private void decompressBundle(long bundleId, File bundleStore) 
    throws IOException
    {
        JarFile jarFile = platform.getBundleJar(bundleId);              
        Enumeration<JarEntry> enumeration = jarFile.entries();
        byte[] bytes = new byte[1024];
        File file = null;
        File directory = null;
        
        while (enumeration.hasMoreElements()) 
        {    
             JarEntry entry = enumeration.nextElement();
             
             file = new File(bundleStore.getAbsolutePath() +
                     File.separatorChar + entry.getName());
             
             if (entry.isDirectory()) 
             { 
                file.mkdirs();
                
             } else 
             {
                 String entryName = entry.getName();
                 String prefix  = entryName.substring(0,entryName.lastIndexOf('/'));                  
                 directory = new File(bundleStore.getAbsolutePath() +
                     File.separatorChar + prefix);

                 if(!directory.exists())
                 {
                     directory.mkdirs();
                 }                
                 InputStream inputStream = jarFile.getInputStream(entry); 
                 FileOutputStream outputStream = new FileOutputStream(file);             
                 while (inputStream.available() > 0) 
                 { 
                   inputStream.read(bytes);
                   outputStream.write(bytes);
                 }
                 outputStream.close();
                 inputStream.close();
             }
        }        
    }
    
    /**
     * Delete recursively all the files contained in the specified file if the
     * specified file is a directory, and the specified file itself.
     * 
     * @param file
     *            the file to delete
     */
    public static void deleteAllRecursively(File file)
    {
        if (file != null && file.exists())
        {
            if (file.isDirectory())
            {
                File[] subs = file.listFiles();
                for (int i = 0; i < subs.length; i++)
                {
                    deleteAllRecursively(subs[i]);
                }
            }
            file.delete();
        }
    }
}
